"""
Module to handle the split of Wyckoff positions
"""
from copy import deepcopy
import numpy as np
from numpy.random import Generator
from pymatgen.core.operations import SymmOp
import pyxtal.symmetry as sym

class wyckoff_split:
    """
    Class for performing wyckoff split between two space groups.
    Essentially, this code is to look for the database from the
    international crystallographic table and find the group-subgroup
    relations

    Args:
        G (int): 1-230, number of super space group or object
        idx (int): index of splitting scheme, default None
        wp1: string ("4a") or integer (1)
        group_type (string): 't' or 'k'
        elements: corresponding chemical species for each wp
    """

    def __init__(self, G=197, idx=None, wp1=None, group_type="t",
                 elements=None, random_state=None):

        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        if wp1 is None:
            wp1 = [0, 1]
        self.error = False
        self.elements = elements
        if type(G) in [int, np.int64]:
            self.G = sym.Group(G)  # Group object
        else:
            self.G = G
        self.group_type = group_type
        if group_type == "t":
            self.wyc = self.G.get_max_t_subgroup()
        else:
            self.wyc = self.G.get_max_k_subgroup()
        id_lists = []
        for wp in wp1:
            if isinstance(wp, int):
                id_lists.append(wp)
            else:
                id_lists.append(sym.index_from_letter(wp[-1], self.G))
        self.wp1_indices = id_lists
        self.wp1_lists = [self.G[id] for id in id_lists]  # a WP object

        # choose a random spliting option if idx is not specified
        if idx is None:
            ids = list(range(len(self.wyc["subgroup"])))
            idx = self.random_state.choice(ids)
        # print(G, idx, len(self.wyc['subgroup']))
        H = self.wyc["subgroup"][idx]
        self.H = sym.Group(H)  # Group object

        # print(G, H)
        self.parse_wp2(idx)

        # check if it is a valid split
        # if self.G.lattice_type == self.H.lattice_type:
        #    self.valid_split = False
        #    for wps in self.wp2_lists:
        #        for wp in wps:
        #            rotation = np.array(wp[0].as_dict()['matrix'])[:3,:3]
        #            if np.linalg.matrix_rank(rotation) > 0:
        #                self.valid_split = True
        #                break
        # else:
        self.valid_split = True

        # if self.valid_split:
        self.G1_orbits = []
        self.G2_orbits = []
        self.H_orbits = []
        for i, wp1 in enumerate(self.wp1_lists):
            self.counter = 0
            self.current_wp1_size = len(wp1)
            self.H_orbits.append([wp2.ops for wp2 in self.wp2_lists[i]])
            if group_type == "t":
                G1_orbits, G2_orbits = self.split_t(wp1, self.wp2_lists[i])
            else:
                G1_orbits, G2_orbits = self.split_k(wp1, self.wp2_lists[i])
            self.G1_orbits.append(G1_orbits)
            self.G2_orbits.append(G2_orbits)
            # self.patch()

    def sort(self):
        """
        Sort the orbits by multiplicity
        This is a utility for the use of supergroup search
        """
        muls = np.array([wp1.multiplicity for wp1 in self.wp1_lists])
        ids = np.argsort(muls)
        self.wp1_lists = [self.wp1_lists[id] for id in ids]
        self.wp2_lists = [self.wp2_lists[id] for id in ids]
        self.G1_orbits = [self.G1_orbits[id] for id in ids]
        self.G2_orbits = [self.G2_orbits[id] for id in ids]
        self.H_orbits = [self.H_orbits[id] for id in ids]
        self.elements = [self.elements[id] for id in ids]

    def parse_wp2(self, idx):
        """
        query the wp2 and transformation matrix from the given {G, H, wp1}
        """
        # print(self.wyc['transformation'])
        # subgroup_relations.reverse()
        trans = self.wyc["transformation"][idx]  # ; print("trans", trans)
        subgroup_relations = self.wyc["relations"][idx]  # ; print('subgroup_relations', subgroup_relations)
        subgroup_relations = list(reversed(subgroup_relations))

        self.R = np.zeros([4, 4])
        self.R[:3, :3] += trans[:3, :3]
        self.R[3, 3] = 1
        self.inv_R = np.linalg.inv(self.R)
        inv_t = np.dot(self.inv_R[:3, :3], trans[:, 3].T)
        self.inv_R[:3, 3] = -1 * inv_t.T
        self.R[:3, 3] = trans[:3, 3]
        self.multi = np.linalg.det(self.R[:3, :3])

        wp2_lists = []
        for wp1_index in self.wp1_indices:
            wp2_list = []
            for letter in subgroup_relations[wp1_index]:
                # if letter == '4a': letter = '8c'
                id = sym.index_from_letter(letter[-1], self.H)
                wp2_list.append(self.H[id])
            wp2_lists.append(wp2_list)
        self.wp2_lists = wp2_lists
        self.index = self.wyc["index"][idx]
        self.cosets = self.wyc["cosets"][idx]
        # import sys; sys.exit()

    def split_t(self, wp1, wp2_lists, quadrant=None):
        """
        split the generators in w1 to different w2s for t-subgroup
        """
        if self.counter == 0:
            self.proper_wp1 = []
            [self.proper_wp1.append(np.array(x.as_dict()["matrix"])) for x in wp1]
            self.original_tau_list = [x[:3, 3] for x in self.proper_wp1]
            for k, _x in enumerate(self.original_tau_list):
                for j in range(3):
                    self.original_tau_list[k][j] = self.original_tau_list[k][j] % 1
                self.original_tau_list[k] = self.original_tau_list[k].round(4)
            for k, _x in enumerate(self.proper_wp1):
                self.proper_wp1[k][:3, 3] = self.original_tau_list[k]
                self.proper_wp1[k] = SymmOp(self.proper_wp1[k])

        wp1_generators_visited = []
        wp1_generators = [np.array(wp.as_dict()["matrix"]) for wp in wp1]

        G1_orbits = []
        G2_orbits = []
        factor = max([1, np.linalg.det(self.R)])

        if quadrant is None:
            quadrant = deepcopy(self.inv_R[:3, 3])
            quadrant[np.abs(quadrant) < 1e-5] = 0
            for i in range(3):
                if quadrant[i] >= 0:
                    quadrant[i] = 1
                else:
                    quadrant[i] = -1
        for wp2 in wp2_lists:
            for gen in wp1_generators:
                good_generator = False
                trans_generator = np.matmul(self.inv_R, gen)
                trans_generator[np.abs(trans_generator) < 1e-5] = 0

                for i in range(3):
                    trans_generator[i][3] = trans_generator[i][3] % quadrant[i]
                    if trans_generator[i][3] == 0 and quadrant[i] == -1:
                        trans_generator[i][3] = -1

                g1_orbits = []
                g2_orbits = []

                for i, wp in enumerate(wp2):
                    new_basis_orbit = np.matmul(wp.as_dict()["matrix"], trans_generator)
                    new_basis_orbit[np.abs(new_basis_orbit) < 1e-5] = 0
                    for j in range(3):
                        new_basis_orbit[j, 3] = new_basis_orbit[j, 3] % quadrant[j]
                        if new_basis_orbit[j, 3] == 0 and quadrant[j] == -1:
                            new_basis_orbit[j, 3] = -1

                    old_basis_orbit = np.matmul(self.R, new_basis_orbit)
                    old_basis_orbit[np.abs(old_basis_orbit) < 1e-5] = 0
                    old_basis_orbit[np.abs(old_basis_orbit - 1) < 1e-5] = 1
                    old_basis_orbit[np.abs(old_basis_orbit + 1) < 1e-5] = -1
                    tmp = deepcopy(old_basis_orbit)
                    tmp[3, :] = [0, 0, 0, 1]
                    # print(
                    #     "tracking wp2 orbit",
                    #     i,
                    #     "newbasisorbit",
                    #     SymmOp(new_basis_orbit).as_xyz_str(),
                    #     "oldbasisorbit",
                    #     SymmOp(old_basis_orbit).as_xyz_str(),
                    #     "chosenwyckoff",
                    #     wp.as_xyz_str(),
                    # )
                    # print('transgenerator',SymmOp(trans_generator).as_xyz_str())
                    if i == 0:
                        truth = True
                        if self.counter != 0:
                            tau = tmp[:3, 3]
                            for j in range(3):
                                tau[j] = tau[j] % 1
                            tau = tau.round(4)
                            temporary = deepcopy(tmp)
                            temporary[:3, 3] = tau
                            temporary = SymmOp(temporary)
                            truth = any(temporary == x for x in self.proper_wp1)
                        # print('current gen',SymmOp(gen).as_xyz_str())
                        # print('current new_basis_orbit',SymmOp(new_basis_orbit).as_xyz_str())
                        # print('current state wp2 orbit',wp.as_xyz_str())
                        # print('wp1generated')
                        # [print(SymmOp(x).as_xyz_str()) for x in wp1_generators_visited]
                        # print('not in wp1 visited',not in_lists(tmp, wp1_generators_visited))
                        # print('in wp1 generators',in_lists(tmp, wp1_generators))
                        if not in_lists(tmp, wp1_generators_visited) and in_lists(tmp, wp1_generators) and truth:
                            good_generator = True
                        else:
                            break
                    # to consider PBC
                    # print(SymmOp(old_basis_orbit).as_xyz_str(),'   ',SymmOp(new_basis_orbit).as_xyz_str(),'    ',wp.as_xyz_str())
                    g1_orbits.append(old_basis_orbit)
                    if self.counter >= 1 and in_lists(new_basis_orbit, g2_orbits):
                        good_generator = False
                        break
                    g2_orbits.append(new_basis_orbit)

                if good_generator:
                    temp = []
                    for gen in g1_orbits:
                        if not in_lists(gen, temp, PBC=False):
                            temp.append(gen)
                    if int(len(temp) * factor) >= len(wp2):
                        wp1_generators_visited.extend(temp)
                        g1_orbits = [SymmOp(orbit) for orbit in g1_orbits]
                        g2_orbits = [SymmOp(orbit) for orbit in g2_orbits]
                        # print('G1=')
                        # [print(x.as_xyz_str()) for x in g1_orbits]
                        # print('G2=')
                        # [print(x.as_xyz_str()) for x in g2_orbits]
                        G1_orbits.append(g1_orbits)
                        G2_orbits.append(g2_orbits)

                        break
            try:
                self.check_orbits(g1_orbits, wp2, wp2_lists)
            except:
                if self.counter != 0:
                    quadrants = [
                        [1, 1, 1],
                        [1, 1, -1],
                        [1, -1, 1],
                        [1, -1, -1],
                        [-1, 1, 1],
                        [-1, 1, -1],
                        [-1, -1, 1],
                        [-1, -1, -1],
                    ]
                    quadrant = quadrants[self.counter - 1]
                wp1_generators = wp1_generators[: self.current_wp1_size]
                wp2_translations = []
                for wp2 in wp2_lists:
                    wp = [np.array(x.as_dict()["matrix"]) for x in wp2]
                    rot = [x[:3, :3] for x in wp]
                    tau = [x[:3, 3] for x in wp]
                    translations = [np.array(tau[i]) for i, x in enumerate(rot) if np.array_equal(x, rot[0])]
                    translations = [x - translations[0] for x in translations]
                    wp2_translations.append(translations)
                new_wp1 = []
                for translation_set in wp2_translations:
                    for translation in translation_set:
                        for gen in wp1_generators:
                            orbit = np.matmul(self.inv_R, gen)
                            orbit[np.abs(orbit) < 1e-5] = 0
                            orbit[np.abs(orbit - 1) < 1e-5] = 1
                            orbit[np.abs(orbit + 1) < 1e-5] = -1

                            for i in range(3):
                                if quadrant[i] == 1:
                                    orbit[i][3] += (translation[i]) % 1
                                    orbit[i][3] = orbit[i][3] % 1
                                else:
                                    orbit[i][3] += (translation[i]) % -1
                                    orbit[np.abs(orbit) < 1e-5] = 0
                                    orbit[np.abs(orbit - 1) < 1e-5] = 1
                                    orbit[np.abs(orbit + 1) < 1e-5] = -1
                                    if orbit[i][3] == 0:
                                        orbit[i][3] = -1
                                    elif orbit[i][3] != -1:
                                        orbit[i][3] = orbit[i][3] % -1
                            orbit = np.matmul(self.R, orbit)
                            orbit[np.abs(orbit) < 1e-5] = 0
                            orbit[np.abs(orbit - 1) < 1e-5] = 1
                            orbit[np.abs(orbit + 1) < 1e-5] = -1
                            orbit = SymmOp(orbit)

                            if orbit not in new_wp1:
                                new_wp1.append(orbit)
                self.counter += 1
                if self.counter == 5:
                    self.valid_split = False
                    self.error = True
                    return None, None
                return self.split_t(new_wp1, wp2_lists, quadrant=quadrant)
        return G1_orbits, G2_orbits

    def split_k(self, wp1, wp2_lists, tol=1e-5):
        """
        split the generators in w1 to different w2s for k-subgroup
        """

        wp1_generators = [np.array(wp.as_dict()["matrix"]) for wp in wp1]

        G1_orbits = []
        G2_orbits = []
        quadrant = deepcopy(self.inv_R[:3, 3])
        quadrant[np.abs(quadrant) < tol] = 0  # finds the orientation of the subgroup_basis

        for i in range(3):
            if quadrant[i] >= 0:
                quadrant[i] = 1
            else:
                quadrant[i] = -1

        all_g2_orbits = []
        translations = self.translation_generator()

        # the translation generator provides all the possible ways to translate
        # the starting positions, then they are shifted
        for translation in translations:
            for gen in wp1_generators:  # into the proper orientation
                orbit = np.matmul(self.inv_R, gen)
                orbit[np.abs(orbit) < tol] = 0
                orbit[np.abs(orbit - 1) < tol] = 1
                orbit[np.abs(orbit + 1) < tol] = -1
                for i in range(3):
                    if quadrant[i] == 1:
                        orbit[i][3] += translation[i]
                        orbit[i][3] = orbit[i][3] % 1
                        if np.abs(orbit[i][3] - 1) < tol:
                            orbit[i][3] = 0
                    else:
                        orbit[i][3] += (translation[i]) % -1
                        orbit[i][3] = orbit[i][3] % -1
                        if np.abs(orbit[i][3]) < tol:
                            orbit[i][3] = -1
                all_g2_orbits.append(orbit)

        for wp2 in wp2_lists:
            # final_G2=[]
            temp = np.array(deepcopy(all_g2_orbits))
            temp[np.abs(temp) < tol] = 0
            temp = temp.tolist()

            for j, x in enumerate(temp):
                temp[j] = SymmOp(x)

            for orbit in temp:
                try_match = np.array([np.matmul(x.as_dict()["matrix"], orbit.as_dict()["matrix"]) for x in wp2])
                try_match[np.abs(try_match) < tol] = 0
                try_match[np.abs(try_match - 1) < tol] = 1
                try_match[np.abs(try_match + 1) < tol] = -1

                # print('\norbit\n', orbit)
                # print('\ntry_match\n', try_match)

                for j in range(len(try_match)):
                    for k in range(3):
                        try_match[j][k][3] = try_match[j][k][3] % quadrant[k]
                        if try_match[j][k][3] == 0 and quadrant[k] == -1:
                            try_match[j][k][3] = -1
                try_match = try_match.tolist()

                for j, x in enumerate(try_match):
                    try_match[j] = SymmOp(x)

                if np.any([try_match.count(x) > 1 for x in try_match]):
                    continue

                try:
                    corresponding_positions = [temp.index(x) for x in try_match]
                except:
                    continue

                for index in sorted(corresponding_positions, reverse=True):
                    del all_g2_orbits[index]
                G2_orbits.append(try_match)
                break

        for position in G2_orbits:
            final_G1 = []
            for orbit in position:
                final_G1.append(SymmOp(np.matmul(self.R, orbit.as_dict()["matrix"])))
            G1_orbits.append(final_G1)

        if len(G1_orbits) != len(wp2_lists):
            print("wp2_lists", wp2_lists)
            print("G1_orbits", G1_orbits)
            print("G2_orbits", G2_orbits)
            raise ValueError("inconsistent G1_orbits and wp2_lists")
        else:
            return G1_orbits, G2_orbits

    def translation_generator(self):
        """
        a function to handle the translation during lattice transformation
        """
        modulo = round(np.linalg.det(self.R[:3, :3]))
        inv_rotation = np.array(self.inv_R[:3, :3]) * modulo
        subgroup_basis_vectors = (np.rint(inv_rotation.transpose()).astype(int) % modulo).tolist()

        # remove the [0,0,0] vectors
        translations = [x for x in subgroup_basis_vectors if x != [0, 0, 0]]

        # find the independent vectors
        if len(translations) == 0:
            independent_vectors = [[0, 0, 0]]
        elif len(translations) == 1:
            independent_vectors = translations
        elif len(translations) == 2:
            norm = round(np.linalg.norm(translations[0]) * np.linalg.norm(translations[1]))
            inner_product = np.inner(translations[0], translations[1])
            difference = norm - inner_product
            independent_vectors = [translations[0]] if difference == 0.0 else translations
        else:
            norms = np.rint(
                [
                    np.linalg.norm(translations[i]) * np.linalg.norm(translations[j])
                    for i in range(2)
                    for j in range(i + 1, 3)
                ]
            )
            inner_products = np.array(
                [np.inner(translations[i], translations[j]) for i in range(2) for j in range(i + 1, 3)]
            )
            differences = inner_products - norms
            independent_vectors = [translations[0]]
            if differences[0] != 0.0 and differences[1] == 0.0:
                independent_vectors.append(translations[1])
            elif differences[0] == 0.0 and differences[1] != 0.0:
                independent_vectors.append(translations[2])
            elif differences[0] != 0.0 and differences[1] != 0.0 and differences[2] != 0.0:
                independent_vectors.append(translations[1])
                independent_vectors.append(translations[2])
            elif differences[0] != 0.0 and differences[1] != 0.0 and differences[2] == 0.0:
                independent_vectors.append(translations[1])

        # generate all possible combinations of the independent vectors
        l = len(independent_vectors)
        independent_vectors = np.array(independent_vectors)
        possible_combos = []
        final_translation_list = []

        for i in range(self.index**l):
            possible_combos.append(np.base_repr(i, self.index, padding=l)[-l:])
        for combo in possible_combos:
            combo = np.array([int(x) for x in combo])
            vector = np.array([0.0, 0.0, 0.0])
            for i, scalar in enumerate(combo):
                vector += scalar * independent_vectors[i]
            vector = (vector % modulo / modulo).tolist()
            if vector not in final_translation_list:
                final_translation_list.append(vector)

        return final_translation_list

    def check_orbits(self, g1_orbits, wp2, wp2_lists):
        if len(g1_orbits) < len(wp2):
            s2 = ""
            for wp2 in wp2_lists:
                s2 += wp2.get_label()
                s2 += ", "
            # g, h = self.G.number, self.H.number
            # print("Error between {:d}[{:s}] -> {:d}[{:s}]".format(g, s1, h, s2))
            # print(self.R)
            # print(g1_orbits)
            # import sys; sys.exit()
            raise ValueError("Cannot find the generator for wp2")

    def __str__(self):
        s = f"Wycokff split from {self.G.number:d} to {self.H.number:d}\n"
        for i, wp1 in enumerate(self.wp1_lists):
            s += f"\n{wp1.get_label():s} -> "

            for j, wp2 in enumerate(self.wp2_lists[i]):
                s += f"{wp2.get_label():s}\n"
                g1s = self.G1_orbits[i][j]
                g2s = self.G2_orbits[i][j]
                Hs = self.H_orbits[i][j]
                for g1_orbit, g2_orbit, h_orbit in zip(g1s, g2s, Hs):
                    g1_xyz = g1_orbit.as_xyz_str()
                    g2_xyz = g2_orbit.as_xyz_str()
                    h_xyz = h_orbit.as_xyz_str()
                    s += f"{g1_xyz:30s} -> {g2_xyz:30s} -> {h_xyz:30s}\n"
        return s

    def __repr__(self):
        return str(self)


def in_lists(mat1, mat2, eps=1e-2, PBC=True):
    if len(mat2) == 0:
        return False
    else:
        for mat in mat2:
            if np.array_equal(mat[:3, :3], mat1[:3, :3]):
                diffs = np.abs(mat[:3, 3] - mat1[:3, 3])
                if PBC:
                    diffs -= np.rint(diffs)
                # print("diffs", diffs)
                if (diffs**2).sum() < eps:
                    return True
        return False


if __name__ == "__main__":
    # sp = wyckoff_split(G=14, idx=1, wp1=['2c', '4e'], group_type='t')
    # print(sp)
    for idx in range(4):
        sp = wyckoff_split(G=210, idx=idx, wp1=['8b'], group_type='k')
        sp = wyckoff_split(G=227, idx=idx, wp1=["96g"], group_type="t")
        #print(sp)
        print(sp.R)
        print(sp.error)



"""
Common utlities for global search
"""
import os
import warnings
from time import time
import concurrent.futures

import numpy as np
from random import choice
from ase import units

from pyxtal import pyxtal
from pyxtal.interface.ase_opt import ASE_relax
from pyxtal.interface.charmm import CHARMM
from pyxtal.optimize.benchmark import benchmark
from pyxtal.representation import representation
from pyxtal.symmetry import Group, Hall
from pyxtal.util import ase2pymatgen
from pyxtal.XRD import Similarity, pxrd_refine

warnings.filterwarnings("ignore")


def get_rmsd_with_timeout(matcher, ref_pmg, structure, timeout=10):
    """Calculate RMSD with a timeout
    
    Args:
        matcher: Structure matcher object
        ref_pmg: Reference pymatgen structure
        structure: Structure to compare
        timeout: Maximum execution time in seconds
        
    Returns:
        RMSD value or None if calculation times out
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(matcher.get_rms_dist, ref_pmg, structure)
        try:
            return future.result(timeout=timeout)
        except concurrent.futures.TimeoutError:
            print(f"RMSD calculation timed out after {timeout} seconds")
            return None

def sweep(xtal, comp, c_info, w_dir, job_tag, skip_ani, optimizer, eps=[0.05, -0.02]):
    """
    Check the stability of input xtal based on 5% tension

    Args:
        xtal: pyxtal structure
        comp: composition
        c_info: CHARMM info
        w_dir: working directory
        job_tag: job tag for CHARMM
        skip_ani: skip ANI relaxation
        optimizer: the optimizer function to use
        eps: list of tension and compression factors

    Returns:
        xtal0: the relaxed structure
        eng0: the energy of the relaxed structure
        stable: whether the structure was updated
    """
    N = sum(xtal.numMols)
    comp = xtal.get_1D_comp()
    res = optimizer(xtal, c_info, w_dir, job_tag, skip_ani=skip_ani)
    if res is None: return None

    xtal0, eng0 = res["xtal"], res["energy"]
    cell0 = np.array(xtal.lattice.encode())
    wps = [site.encode() for site in xtal0.mol_sites]
    smiles = [m.smile for m in xtal.molecules]

    # Check if the structure is stable w.r.t. 5% tension/compression
    stable = True
    sg = xtal.group.number
    abc = 3 if sg <= 74 else (2 if sg <= 194 else 1)
    angle = 2 if sg <= 15 else 1
    for id in range(abc):
        for eps_i in eps:
            for ang in range(angle):
                cell = cell0.copy()
                cell[id] *= (1 + eps_i)
                # Apply a small shear strain if the angle is allowed
                if sg <=15:
                    if ang == 0:
                        cell[-1] += 2.0
                    else:
                        cell[-1] -= 2.0
                x = [[xtal0.group.hall_number] + cell.tolist()]
                x.extend(wps)
                rep1 = representation(x, smiles)
                try:
                    xtal1 = rep1.to_pyxtal(composition=comp)
                except:
                    print("Problem in rep.to_pyxtal", rep1)
                    xtal1 = None
                if xtal1 is not None:
                    res = optimizer(xtal1, c_info, w_dir, job_tag, skip_ani=skip_ani)
                    if res is not None:
                        xtal2, eng = res["xtal"], res["energy"]
                        if eng < eng0 - 1e-2:
                            rep2 = xtal2.get_1D_representation()
                            print(rep2.to_string(eng/N), f"<- {eng0/N:.2f} ({id}@{eps_i:.2f}/{ang})")
                            xtal0, eng0 = xtal2, eng
                            stable = False
    return xtal0, eng0, stable


def check_stable_structure(xtal, c_info, w_dir, job_tag, skip_ani, optimizer, disps=[0.5, 5.0], random=False):
    """
    Check the stability of input xtal based on lattice mutation
    """
    N = sum(xtal.numMols)
    comp = xtal.get_1D_comp()
    disp_cell, disp_ang = disps[0], disps[1]
    res = optimizer(xtal, c_info, w_dir, job_tag, skip_ani=skip_ani)
    smiles = [m.smile for m in xtal.molecules]#; print(smiles)
    if res is not None:
        xtal0, eng0 = res["xtal"], res["energy"]
        cell0 = np.array(xtal.lattice.encode())

        for i, c in enumerate(cell0):
            if i <= 2:
                disps = [-disp_cell, disp_cell]
            else:
                disps = [-disp_ang, disp_ang]
            for disp in disps:
                cell = cell0.copy()
                cell[i] += disp
                x = [[xtal0.group.hall_number] + cell.tolist()]
                wps = [site.encode() for site in xtal0.mol_sites]
                # Add random perturbation
                if random:
                    for wp in wps:
                        for id in range(4, len(wp)-1):
                            wp[id] += 10.0*(np.random.random()-0.5)
                x.extend(wps)
                rep1 = representation(x, smiles)
                xtal1 = rep1.to_pyxtal(composition=comp)#; print(xtal1.lattice)
                res = optimizer(xtal1, c_info, w_dir, job_tag, skip_ani=skip_ani)
                if res is not None:
                    xtal2, eng = res["xtal"], res["energy"]
                    if eng < eng0 + 1e-4:
                        xtal0, eng0 = xtal2, eng
        return xtal0, eng0
    else:
        #raise RuntimeError("Error in optimization")
        return None


def mutator(xtal, smiles, opt_lat, ref_pxrd=None, dr=0.125, random_state=None):
    """
    A random mutation.

    Args:
        xtal: pyxtal structure
        smiles: list of smiles
        opt_lat: whether to optimize the lattice
        ref_pxrd: reference PXRD pattern for lattice optimization
        dr: perturbation ratio for the lattice and molecules
        random_state: random state for reproducibility

    Returns:
        a new pyxtal structure with perturbed lattice and molecules
    """

    rng = np.random.default_rng(random_state)
    # perturb cell
    comp = xtal.get_1D_comp()
    x = xtal.get_1D_representation().x
    if opt_lat:
        if ref_pxrd is None:
            sg = x[0][0]
            if rng.random() > 0.5:
                disp_cell = rng.uniform(-1.0, 1.0, len(x[0]) - 1)
                if sg <= 2: disp_cell[3:] = 0 # prevent some bad angles
                x[0][1:] *= 1 + dr * disp_cell

                # flip the inclination angle
                if 3 <= sg <= 15 and rng.random() > 0.7 and abs(90 - x[0][-1]) < 15:
                    x[0][-1] = 180 - x[0][-1]
            else:
                # pure tension 15-25%
                if sg <= 74: # pick a, b, c
                    axis = rng.choice([0, 1, 2])
                elif sg <= 194: # a, c
                    axis = rng.choice([0, 1])
                else: # c
                    axis = 0
                x[0][axis] *= 1.15 + 0.1 * rng.random()
        else:
            thetas = [ref_pxrd[0][0], min([35, ref_pxrd[0][-1]])]
            _, x[0][1:], _ = pxrd_refine(xtal, ref_pxrd, thetas)

    # perturb molecules
    for i in range(1, len(x)):
        disp_mol = rng.uniform(-1.0, 1.0, len(x[i]) - 1)
        x[i][:-1] *= 1 + dr * disp_mol
        # change the orientation and torsions
        for j in range(3, len(x[i]) - 1):
            rad_num = rng.random()
            if rad_num < 0.25:
                x[i][j] += rng.choice([45.0, 90.0])
            elif rad_num < 0.5:
                x[i][j] *= -1
    try:
        struc = representation(x, smiles).to_pyxtal(composition=comp)
        for i, molecule in enumerate(xtal.molecules):
            if hasattr(molecule, 'active_sites'):
                struc.molecules[i].active_sites = molecule.active_sites
        return struc
    except:
        #print("Trouble to get the representation")
        #print(xtal)
        #print("x", x)
        #print("smiles", smiles)
        #print("comp", comp)
        #xtal.to_file("bug.cif")
        ##print("is_valid_matrix\n", xtal.lattice.get_matrix())
        #print("cell_para", xtal.lattice.get_para(degree=True))
        #print(x[0])
        #raise RuntimeError("Problem occurs in mutation_lattice")
        return None

def randomizer(
    smiles,
    sgs,
    comp,
    lattice=None,
    block=None,
    num_block=None,
    torsions=None,
    molecules=None,
    sites=None,
    use_hall=False,
    factor=1.1,
    random_state=None,
):
    """
    A random structure generation engine

    Args:
        smiles: e.g. `['CCCCC', 'CC']`
        sgs: e.g. `[2, 4, 14]`
        comp: e.g. `[0.5, 0.5]`
        lattice: pyxtal.Lattice object
        block: a list of block sizes, e.g. `[2, 2, 2]`
        num_block: number of blocks, e.g. `2`
        torsions: a list of torsion angles, e.g. `[0, 90, 180]`
        molecules: pre-specified pyxtal_molecule object
        sites: a list of sites, e.g. `[['1a', '1b'], ['2c']]`
        use_hall: whether to use Hall notation
        factor: factor to scale the lattice
        random_state: random state for reproducibility

    Returns:
        PyXtal object
    """
    rng = np.random.default_rng(random_state)
    mols = [smi + ".smi" for smi in smiles] if molecules is None else [choice(m) for m in molecules]
    sg = rng.choice(sgs)
    wp = Group(sg, use_hall=True)[0] if use_hall else Group(sg)[0]
    mult = len(wp)
    numIons = [int(c * mult) for c in comp]

    # speed up generation for general positions
    if sites is None and comp[0] >= 1:
        letter = wp.letter
        sites = []
        for c in comp:
            sites.append([str(mult) + letter] * int(c))

    while True:
        xtal = pyxtal(molecular=True)
        if use_hall:
            hn = sg
        else:
            perm = sg > 15
            # For specical site or high symmetry, we only do standard_setting
            hn = Hall(sg).hall_default if min(comp) < 1 or sg > 142 else rng.choice(Hall(sg, permutation=perm).hall_numbers)
        xtal.from_random(
            3,
            hn,
            mols,
            numIons,
            factor=factor,
            block=block,
            num_block=num_block,
            lattice=lattice,
            force_pass=True,
            torsions=torsions,
            sites=sites,
            use_hall=True,
            #random_state=random_state,
        )
        if xtal.valid and len(xtal.check_short_distances(exclude_H=True))==0:
            break

    if xtal.has_special_site():
        try:
            xtal = xtal.to_subgroup()
        except:
            print(xtal)
            print(xtal.to_file())
            raise ValueError("Error in making subgroup")
    return xtal


def optimizer(
    struc,
    atom_info,
    workdir,
    tag = "job_0",
    opt_lat = True,
    calculators = None,
    max_time = 180,
    skip_ani = False,
    pre_opt = False,
):
    """
    Structural relaxation for each individual pyxtal structure.

    Args:
        struc: pyxtal
        atom_info: atom information for CHARMM
        workdir: working directory
        tag: job tag for CHARMM
        opt_lat: whether to optimize the lattice
        calculators: e.g., `['CHARMM', 'GULP']`
        max_time: maximum time for the optimization
        skip_ani: whether to skip ANI relaxation
        pre_opt: whether to perform pre-relaxation

    Returns:
        a dictionary with xtal, energy and time
    """
    # Perform pre-relaxation
    if pre_opt:
        struc.optimize_lattice_and_rotation()

    if calculators is None:
        calculators = ["CHARMM"]

    cwd = os.getcwd()
    t0 = time()
    os.chdir(workdir)

    results = None
    for i, calculator in enumerate(calculators):
        if calculator == "CHARMM":
            if i == 0:
                calc = CHARMM(struc, tag, steps=[1000], atom_info=atom_info)#, debug=True)
                calc.run() #clean=False); import sys; sys.exit()
                if calc.error:
                    os.chdir(cwd)
                    return None
                else:
                    if not calc.optimized:
                        calc = CHARMM(struc, tag, steps=[500], atom_info=atom_info)
                        calc.run()
                        if calc.error:
                            os.chdir(cwd)
                            return None

            # print(calc.structure)
            steps = [1000, 1000] if opt_lat else [2000]
            calc = CHARMM(calc.structure, tag, steps=steps, atom_info=atom_info)
            calc.run()  # print("Debug", calc.optimized); import sys; sys.exit()

            # only count good struc
            if calc.error:
                os.chdir(cwd)
                return None
            else:
                calc = CHARMM(calc.structure, tag, steps=steps, atom_info=atom_info)
                calc.run()#clean=False)

                if calc.error:
                    os.chdir(cwd)
                    return None
                else:
                    if calc.optlat:  # with bad inclination angles
                        calc = CHARMM(calc.structure, tag, steps=steps, atom_info=atom_info)
                        calc.run()#clean=False)
                        if calc.error:
                            os.chdir(cwd)
                            return None

                # Check if there exists a 2nd FF model for better energy ranking
                if os.path.exists("pyxtal1.prm"):
                    calc = CHARMM(
                        calc.structure,
                        tag,
                        prefix="pyxtal1",
                        steps=[2000],
                        atom_info=atom_info,
                    )
                    calc.run()
                    if calc.error:
                        os.chdir(cwd)
                        return None

        if calc.error:
            os.chdir(cwd)
            return None
        else:
            struc = calc.structure
            struc.resort()
    os.chdir(cwd)

    # density should not be too small
    if not skip_ani:
        stress_tol = 10.0 if len(struc.mol_sites[0].molecule.mol) < 10 else 5.0
        if (
            struc.energy < 9999
            and struc.lattice.is_valid_matrix()
            # and struc.check_distance()
            and 0.25 < struc.get_density() < 3.0
        ):
            s = struc.to_ase()
            s = ASE_relax(s, 'ANI', step=50, fmax=0.1, logfile="ase.log")
            if s is None: return None
            eng = s.get_potential_energy()
            stress = max(abs(s.get_stress())) / units.GPa

            t = time() - t0
            if t > max_time:
                try:
                    print("!!!Long time in ani calculation", t)
                    print(struc.get_1D_representation().to_string())
                    struc.optimize_lattice()
                except:
                    print("Trouble in optLat")
                    return None
            elif stress < stress_tol:
                results = {}
                results["xtal"] = struc
                if eng is not None:
                    results["energy"] = eng
                else:
                    results["energy"] = 10000
                    return None
                results["time"] = time() - t0
            else:
                print(f"stress is wrong {stress:6.2f}")
                # print(struc); import sys; sys.exit()
                return None
    else:
        results = {}
        results["xtal"] = struc
        results["energy"] = calc.structure.energy
        results["time"] = time() - t0

    return results

def optimizer_par(
    xtals,
    ids,
    mutates,
    job_tags,
    randomizer,
    optimizer,
    smiles,
    block,
    num_block,
    atom_info,
    workdir,
    sgs,
    comp,
    lattice,
    torsions,
    molecules,
    sites,
    ref_pmg,
    matcher,
    ref_pxrd,
    use_hall,
    skip_ani,
    check_stable,
    pre_opt,
):
    """
    A routine used for parallel structure optimization

    Args:
        xtals: list of xtals
        ids: list of structure ids
    """
    results = []
    for i in range(len(ids)):
        xtal = xtals[i]
        id = ids[i]
        mutate = mutates[i]
        job_tag = job_tags[i]
        xtal, match, stable = optimizer_single(
            xtal,
            id,
            mutate,
            job_tag,
            randomizer,
            optimizer,
            smiles,
            block,
            num_block,
            atom_info,
            workdir,
            sgs,
            comp,
            lattice,
            torsions,
            molecules,
            sites,
            ref_pmg,
            matcher,
            ref_pxrd,
            use_hall,
            skip_ani,
            check_stable,
            pre_opt,
        )
        results.append((id, xtal, match, stable))
    return results


def optimizer_single(
    xtal,
    id,
    mutate,
    job_tag,
    randomizer,
    optimizer,
    smiles,
    block,
    num_block,
    atom_info,
    workdir,
    sgs,
    comp,
    lattice,
    torsions,
    molecules,
    sites,
    ref_pmg,
    matcher,
    ref_pxrd,
    use_hall,
    skip_ani,
    check_stable,
    pre_opt,
):
    """
    A routine used for individual structure optimization

    Args:
        xtal:
        id: structure id
        randomizer:
        optimizer:
    """

    # 1. Obtain the structure model
    opt_lat = lattice is None
    if xtal is None:
        xtal = randomizer(
            smiles,
            sgs,
            comp,
            lattice,
            block,
            num_block,
            torsions,
            molecules,
            sites,
            use_hall,
        )
        tag = "Random  "
    else:
        if mutate:
            xtal = mutator(xtal, smiles, opt_lat, None)
            tag = "Mutation "
        else:
            tag = "Kept "

    # 2. Optimization
    if xtal is None:
        res = None
    else:
        res = optimizer(xtal, atom_info, workdir, job_tag, opt_lat,
                        skip_ani=skip_ani, pre_opt=pre_opt)

    match = False # used for matching with reference
    stable = True # used for tagging if the structure is stable
    if res is not None:
        xtal, eng = res["xtal"], res["energy"]
        N = sum(xtal.numMols)
        if check_stable and eng < 9999.:
            res = sweep(xtal, comp, atom_info, workdir, job_tag,
                        skip_ani, optimizer)
            if res is not None:
                xtal, eng, stable = res
                if stable:
                    tag += 'Stable'
                else:
                    tag += 'Shallow'
        rep = xtal.get_1D_representation()
        strs = rep.to_string(None, eng / N, tag)

        # 3. Check match w.r.t the reference
        if ref_pmg is not None:
            pmg_s1 = xtal.to_pymatgen()
            pmg_s1.remove_species("H")
            # Prevent the false call of matcher call
            try:
                rmsd = get_rmsd_with_timeout(matcher, ref_pmg, pmg_s1)
            except:
                rmsd = None
            if rmsd is not None:
                # Further refine the structure
                match = True
                str1 = f"Match {rmsd[0]:6.2f} {rmsd[1]:6.2f} {eng / N:12.3f} "
                if not skip_ani:
                    xtal, eng1 = refine_struc(xtal, smiles, ASE_relax)
                    str1 += f"Full Relax -> {eng1 / N:12.3f}"
                    eng = eng1
                print(str1)

        elif ref_pxrd is not None:
            thetas = [ref_pxrd[0][0], ref_pxrd[0][-1]]
            xrd = xtal.get_XRD(thetas=thetas)
            p1 = xrd.get_profile(res=0.15, user_kwargs={"FWHM": 0.25})
            match = Similarity(p1, ref_pxrd, x_range=thetas).value
            strs += f" {match:.3f}"

        xtal.energy = eng
        print(f"{id:3d} " + strs)#; import sys; sys.exit()
        return xtal, match, stable
    else:
        return None, match, stable

def refine_struc(xtal, smiles, calculator):
    """
    refine the structure with the ML calculator

    Args:
        xtal: pyxtal structure
        smiles: list of smiles
        calculator: ANI_relax or MACE_relax
    """
    s = xtal.to_ase()
    s = calculator(s, 'ANI', step=50, fmax=0.1, logfile="ase.log")
    s = calculator(s, 'ANI', step=250, opt_cell=True, logfile="ase.log")
    s = calculator(s, 'ANI', step=50, fmax=0.1, logfile="ase.log")
    eng1 = s.get_potential_energy()  # /sum(xtal.numMols)

    xtal = pyxtal(molecular=True)
    pmg = ase2pymatgen(s)
    mols = [smi + ".smi" for smi in smiles]
    xtal.from_seed(pmg, mols)
    return xtal, eng1

def compute_par(row, pmg, work_dir, skf_dir, queue, compute):
    """
    Args:
        row: ase db row object
        pmg: pymatgen structure
        work_dir: working directory
        skf_dir: directory for the skf files
        queue: multiprocessing.Queue to store results
    """
    data = compute(row, pmg, work_dir, skf_dir)
    queue.put((row.id, row.csd_code, data))

def compute(row, pmg, work_dir, skf_dir, info=None):
    """
    perform the benchmark for a ase db row object

    Arg:
        row: ase db
    """
    data = {
        "charmm_info": {},
        "gulp_info": {},
        "ani_info": {},
        "dftb_D3_info": {},
        "dftb_TS_info": {},
    }

    if info is None:
        c_info = row.data["charmm_info"]
        g_info = row.data["gulp_info"]
    else:
        c_info = info["c_info"]
        g_info = info["g_info"]
    data["charmm_info"] = c_info
    data["gulp_info"] = g_info

    w_dir = work_dir + "/" + row.csd_code
    smiles = row.mol_smi.split(".")

    # Invoke benchmark
    ben = benchmark(
        pmg,
        smiles,
        work_dir=w_dir,
        skf_dir=skf_dir,
        charmm_info=c_info,
        gulp_info=g_info,
    )
    file1, file2 = "/pyxtal.prm", "/pyxtal.rtf"
    with open(w_dir + file1, "w") as prm:
        prm.write(c_info["prm"])
    with open(w_dir + file2, "w") as rtf:
        rtf.write(c_info["rtf"])

    # reference
    print("\n", row.csd_code, row.mol_smi, pmg.volume)
    rep = representation(ben.rep["reference"], smiles)
    print(rep.to_string() + " reference")

    for calc in ["charmm", "gulp", "ani", "dftb_D3", "dftb_TS"]:
        try:
            ben.calc(calc, show=True)
            key = calc + "_info"
            data[key]["rep"] = ben.rep[calc]
            data[key]["time"] = ben.time[calc]
            data[key]["diff"] = ben.diff[calc]
            data[key]["cif"] = ben.xtal[calc].to_file()
            data[key]["energy"] = ben.energy[calc] / sum(ben.xtal[calc].numMols)
        except:  # noqa: PERF203
            print("=====Calculation is wrong ", calc, row.csd_code)
            data[key]["energy"] = 100000

    return data

def load_reference_from_db(db_name, code=None):
    """
    Load the reference data from the db file

    Args:
        db_name (str): database path
        code (str): code name

    Returns:
        a list of args
    """
    from pyxtal.db import database
    db = database(db_name)
    if code is None:
        codes = db.codes
    else:
        codes = [code]

    args = []
    for code in codes:
        wdir = code
        row = db.get_row(code)
        xtal = db.get_pyxtal(code)
        smile, wt, spg = row.mol_smi, row.mol_weight, row.space_group.replace(" ", "")

        chm_info = None
        if 'charmm_info' in row.data.keys():
            print("prepare charmm input", wdir)
            os.makedirs(wdir, exist_ok=True)
            os.makedirs(wdir+'/calc', exist_ok=True)

            chm_info = row.data['charmm_info']
            prm_file = wdir+'/calc/pyxtal.prm'
            rtf_file = wdir+'/calc/pyxtal.rtf'
            prm = open(prm_file, 'w'); prm.write(chm_info['prm']); prm.close()
            rtf = open(rtf_file, 'w'); rtf.write(chm_info['rtf']); rtf.close()

        if xtal.has_special_site(): xtal = xtal.to_subgroup()

        pmg0 = xtal.to_pymatgen()
        sg = xtal.group.number
        comp = xtal.get_zprime(integer=True)
        N_torsion = xtal.get_num_torsions()
        lat = xtal.lattice
        tag = code.lower()
        args.append((smile, wdir, sg, tag, chm_info, comp, lat, pmg0, wt, spg, N_torsion))
    return args

if __name__ == "__main__":

    import pymatgen.analysis.structure_matcher as sm
    from pyxtal.db import database

    w_dir = "tmp"
    if not os.path.exists(w_dir):
        os.makedirs(w_dir)

    db = database("pyxtal/database/test.db")
    row = db.get_row("ACSALA")
    xtal1 = db.get_pyxtal("ACSALA")
    smile = row.mol_smi

    # Prepare prm
    c_info = row.data["charmm_info"]
    with open(w_dir + "/pyxtal.prm", "w") as prm:
        prm.write(c_info["prm"])
    with open(w_dir + "/pyxtal.rtf", "w") as rtf:
        rtf.write(c_info["rtf"])

    # Relax expt. xtal
    res = optimizer(xtal1, c_info, w_dir)
    xtal1, eng = res["xtal"], res["energy"]
    rep0 = xtal1.get_1D_representation()
    print(rep0.to_string(eng))

    # Redo it from the 1D. Rep.
    rep2 = representation.from_string(rep0.to_string(), [smile])
    xtal2 = rep2.to_pyxtal()
    res = optimizer(xtal2, c_info, w_dir)
    xtal2, eng = res["xtal"], res["energy"]
    rep0 = xtal2.get_1D_representation()
    print(rep0.to_string(eng))

    pmg1 = xtal1.to_pymatgen()
    pmg2 = xtal2.to_pymatgen()
    pmg1.remove_species("H")
    pmg2.remove_species("H")
    print(sm.StructureMatcher().fit(pmg1, pmg2))

    reps = [
    "81 9.71  6.19 14.25  84.7 1 0 0.21 0.44 0.12 169.6 -16.9  176.2   77.6  9.6   24.9 0",
    "81 8.38 10.06 11.10 107.8 1 0 0.26 0.42 0.31 118.8 -22.6 -111.9 -117.3  0.4   11.9 0",
    "82 9.37  7.92 12.13 111.0 1 0 0.29 0.34 0.10 155.5 -27.6 -161.1   74.6 10.7 -149.8 0",
    ]
    for rep in reps:
        rep = representation.from_string(rep, [smile])
        xtal1 = rep.to_pyxtal()
        check_stable_structure(xtal1, c_info, w_dir, skip_ani=True, optimizer=optimizer)
"""
 81 11.38 6.48 11.24 96.9 1 0 0.23 0.43 0.03 -44.6 25.0 34.4 -76.6 -5.2 171.5 0 -70594.48
 81 11.38 6.48 11.24 96.9 1 0 0.23 0.43 0.03 -44.6 25.0 34.4 -76.6 -5.2 171.5 0 -70594.48
True
"""



import signal
from time import time
import numpy as np
from ase.constraints import FixSymmetry
from ase.filters import UnitCellFilter
from ase.optimize.fire import FIRE
import logging

import os
from mace.calculators import mace_mp
_cached_mace_mp = None


def get_calculator(calculator):
    global _cached_mace_mp

    if type(calculator) is str:
        if calculator == 'ANI':
            import torchani
            calc = torchani.models.ANI2x().ase()
        else:
            if _cached_mace_mp is None:
                _cached_mace_mp = mace_mp(model='small',
                                          dispersion=True,
                                          device='cpu')
            calc = _cached_mace_mp
    else:
        calc = calculator

    return calc

def ASE_relax(struc, calculator, opt_cell=False, step=500,
              fmax=0.1, logfile=None, max_time=10.0,
              label='ase'):
    """
    ASE optimizer
    Args:
        struc: ase atoms object
        calculator (str): 'ANI', 'MACE'
        opt_cell (bool): optimize cell or not
        step (int): optimization steps
        fmax: float (eV/Angstrom), force convergence criteria
        logfile: str, output file
        max_time: float (minutes)
    """

    def handler(signum, frame):
        raise TimeoutError("Optimization timed out")

    step_init = min([30, int(step/2)])
    logger = logging.getLogger()
    max_time *= 60
    timeout = int(max_time)
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout)
    #logger.info(f"{label} start calculation")
    _fmax = 1e+5

    try:
    #if True:
        calc = get_calculator(calculator)
        struc.set_calculator(calc)
        struc.set_constraint(FixSymmetry(struc))
        if opt_cell:
            ecf = UnitCellFilter(struc)
            dyn = FIRE(ecf, a=0.1, logfile=logfile) if logfile is not None else FIRE(ecf, a=0.1)
        else:
            dyn = FIRE(struc, a=0.1, logfile=logfile) if logfile is not None else FIRE(struc, a=0.1)

        # Run relaxation
        dyn.run(fmax=fmax, steps=step_init)
        forces = dyn.optimizable.get_forces()
        _fmax = np.sqrt((forces ** 2).sum(axis=1).max())

        if _fmax < 1e+3 and step > step_init:
            dyn.run(fmax=fmax, steps=step-step_init)
            forces = dyn.optimizable.get_forces()
            _fmax = np.sqrt((forces ** 2).sum(axis=1).max())
            if _fmax > 100:
                #eng = struc.get_potential_energy() / len(struc)
                #logger.info(f"Warning {label} big stress {eng:.2f} / {_fmax:.2f}, skip")
                struc = None
            #else:
            #    logger.info(f"{label} Success  {eng:.2f} / {_fmax:.2f}")
        else:
            #logger.info(f"Warning {label} big stress {_fmax:.2f} for 20 steps, skip")
            struc = None
        signal.alarm(0)  # Cancel the alarm if finished within time

    except TimeoutError:
        logger.info(f"Warning {label} timed out after {timeout} seconds.")
        struc = None

    except TypeError:
        logger.info(f"Warning {label} spglib error in getting the lattice")
        struc = None
        signal.alarm(0)  # Cancel the alarm if finished within time

    # tag = 'False' if struc is None else 'True'
    # logger.info(f"Finishing {label} {tag}")
    # signal.alarm(0)  # Cancel the alarm
    return struc #, eng, _fmax

class ASE_optimizer:
    """
    This is a ASE optimizer to perform oragnic crystal structure optimization.
    We assume that the geometry has been well optimized by classical FF.

    Args:
        struc: pyxtal object
        calculator (str): 'ANI', 'MACE'
        opt_lat (bool): to opt lattice or not
        log_file (str): output file
    """

    def __init__(self, struc, calculator='MACE', opt_lat=True, logfile=None):
        self.structure = struc
        self.calculator = get_calculator(calculator)
        self.opt_lat = opt_lat
        self.stress = None
        self.forces = None
        self.optimized = True
        self.positions = None
        self.cell = None
        self.cputime = 0
        self.logfile = logfile

    def run(self, steps=10):
        t0 = time()
        s = self.structure.to_ase(resort=False)
        s.set_constraint(FixSymmetry(s))
        s.set_calculator(self.calculator)#; print("Setup Fire")

        if not self.opt_lat:
            dyn = FIRE(s, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
        else:
            #ecf = FrechetCellFilter(s)
            ecf = UnitCellFilter(s)
            dyn = FIRE(ecf, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
            self.structure.lattice.set_matrix(s.get_cell())

        positions = s.get_scaled_positions()
        #try:
        if True:
            # s.write('../1.cif', format='cif')
            count = 0
            for _i, site in enumerate(self.structure.mol_sites):
                coords0, _ = site._get_coords_and_species(first=True)
                coords1 = positions[count : count + len(site.molecule.mol)]
                for j, coor in enumerate(coords1):
                    diff = coor - coords0[j]
                    diff -= np.round(diff)
                    abs_diff = np.dot(diff, s.get_cell())
                    # print(j, coor, coords0[j], diff, np.linalg.norm(abs_diff))
                    if abs(np.linalg.norm(abs_diff)) < 2.0:
                        coords1[j] = coords0[j] + diff
                    else:
                        print(coords1[j], coords1[j], np.linalg.norm(abs_diff))
                        import sys; sys.exit()

                site.update(coords1, self.structure.lattice)
                count += len(site.molecule.mol) * site.wp.multiplicity
            self.structure.optimize_lattice()
            self.structure.energy = s.get_potential_energy()
            self.cell = s.get_cell()
            # print(self.structure.lattice)
        #except:
        #    self.structure.energy = 10000
        #    self.optimized = False
        #    print("Structure is wrong after optimization")
        s.set_calculator()
        s.set_constraint()
        self.cputime = time() - t0


if __name__ == "__main__":
    import os, warnings
    from pyxtal.db import database
    warnings.filterwarnings("ignore")

    work_dir = "tmp"
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    db = database("pyxtal/database/test.db")
    struc = db.get_pyxtal("ACSALA")

    calc = ASE_optimizer(struc)
    print(calc.structure.lattice)
    calc.run()
    print(calc.structure.energy)
    print(calc.structure.lattice)
